#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from concurrent import futures
import datetime
import multiprocessing

from oslo_config import cfg
from oslo_log import log as logging

from octavia.db import api as db_api
from octavia.db import repositories as repo

from a10_octavia.controller.worker import controller_worker as cw
from a10_octavia.db import repositories as a10repo

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
mp_mgr = multiprocessing.Manager()
hk_ctx_map = mp_mgr.dict()
hk_ctx_lock = mp_mgr.Lock()


class SpareAmphora(object):
    def __init__(self):
        self.vthunder_repo = a10repo.VThunderRepository()
        self.cw = cw.A10ControllerWorker()

    def spare_check(self):
        """Checks the DB for the Spare amphora count.

        If it's less than the requirement, starts new amphora.
        """
        session = db_api.get_session()
        conf_spare_cnt = CONF.a10_house_keeping.spare_amphora_pool_size
        curr_spare_cnt = self.vthunder_repo.get_spare_vthunder_count(session)
        busy_spare_cnt = self.vthunder_repo.get_busy_spare_vthunder_count(session)
        curr_spare_cnt = curr_spare_cnt + busy_spare_cnt
        LOG.debug("Required Spare vThunder count : %d", conf_spare_cnt)
        LOG.debug("Current Spare vThunder count : %d", curr_spare_cnt)
        diff_count = conf_spare_cnt - curr_spare_cnt

        # When the current spare amphora is less than required
        if diff_count > 0:
            LOG.info("Initiating creation of %d spare amphora.", diff_count)

            # Call Amphora Create Flow diff_count times
            with futures.ThreadPoolExecutor(
                    max_workers=CONF.a10_house_keeping.spare_amphora_pool_size
            ) as executor:
                for i in range(1, diff_count + 1):
                    LOG.debug("Starting amphorae number %d ...", i)
                    executor.submit(self.cw.create_amphora)
        else:
            LOG.debug("Current spare vThunder count satisfies the requirement")


class DatabaseCleanup(object):
    def __init__(self):
        self.vthunder_repo = a10repo.VThunderRepository()
        self.lb_repo = a10repo.LoadBalancerRepository()

    def delete_old_amphorae(self):
        """Checks the DB for old amphora and deletes them based on its age."""
        exp_age = datetime.timedelta(
            seconds=CONF.a10_house_keeping.amphora_expiry_age)

        session = db_api.get_session()
        amp_ids = self.vthunder_repo.get_all_deleted_expiring(session,
                                                              exp_age=exp_age)
        LOG.info('VThunder ids: %s', amp_ids)

        for amp_id in amp_ids:
            LOG.info('Attempting to purge db record for VThunder ID: %s',
                     amp_id)
            self.vthunder_repo.delete(session, id=amp_id)
            LOG.info('Purged db record for Amphora ID: %s', amp_id)

    def cleanup_load_balancers(self):
        """Checks the DB for old load balancers and triggers their removal."""
        exp_age = datetime.timedelta(
            seconds=CONF.a10_house_keeping.load_balancer_expiry_age)

        session = db_api.get_session()
        lb_ids = self.lb_repo.get_all_deleted_expiring(session,
                                                       exp_age=exp_age)
        LOG.info('Load balancer ids: %s', lb_ids)
        for lb_id in lb_ids:
            LOG.info('Attempting to delete load balancer id : %s', lb_id)
            self.lb_repo.delete(session, id=lb_id)
            LOG.info('Deleted load balancer id : %s', lb_id)


class StatisticsCleanup(object):
    def __init__(self):
        self.listener_repo = repo.ListenerRepository()
        self.listener_stats_repo = a10repo.ListenerStatisticsRepository()

    def delete_listener_stats(self):

        listeners, _ = self.listener_repo.get_all(db_api.get_session())
        listeners_stats, _ = self.listener_stats_repo.get_all(db_api.get_session())

        listeners_ids = [listener.id for listener in listeners]
        listeners_stats_ids = [listener_stats.listener_id for listener_stats in listeners_stats]

        for listeners_stats_id in listeners_stats_ids:
            if listeners_stats_id not in listeners_ids:
                self.listener_stats_repo.delete_multiple(db_api.get_session(),
                                                         listener_id=listeners_stats_id)
                LOG.info('Delete the statictics entry of listener %s', listeners_stats_id)


class WriteMemory(object):

    def __init__(self):
        self.thunder_repo = a10repo.VThunderRepository()
        self.cw = cw.A10ControllerWorker()
        self.svc_up_time = datetime.datetime.utcnow()

    def perform_memory_writes(self):
        thunders = self.thunder_repo.get_recently_updated_thunders(db_api.get_session())
        ip_partition_list = set()
        write_mem_list = []
        reload_check_list = []
        for thunder in thunders:
            ip_partition = str(thunder.ip_address) + ":" + str(thunder.partition_name)
            if (thunder.status != 'DELETED' and
                    thunder.loadbalancer_id is not None):
                if thunder.last_write_mem is not None:
                    reload_check_list.append(thunder)
                elif thunder.updated_at > self.svc_up_time:
                    # For new lb, it didn't have last_write_mem yet, so if
                    #  - it's latest update is after this service starts, then
                    #    some config is not save yet.
                    #  - it's latest update time is before this service starts, then
                    #    we can't make sure it has config not save or not. Because
                    #    periodical write memory may not enabled at that time.
                    reload_check_list.append(thunder)
            if ip_partition not in ip_partition_list:
                ip_partition_list.add(ip_partition)
                write_mem_list.append(thunder)

        if reload_check_list:
            LOG.info("Check configuration lost for Thunders : %s", list(reload_check_list))
            self.cw.perform_reload_check(reload_check_list)
            LOG.info("Finished conf lost check for {} thunders...".format(len(reload_check_list)))

        if write_mem_list:
            LOG.info("Write Memory for Thunders : %s", list(ip_partition_list))
            self.cw.perform_write_memory(write_mem_list)
            LOG.info("Finished running write memory for {} thunders...".format(len(write_mem_list)))
        else:
            LOG.warning("No thunders found that are recently updated."
                        " Not performing write memory...")


class PendingResourceCleanup(object):

    def __init__(self):
        self.loadbalancer_repo = a10repo.LoadBalancerRepository()
        self.vthunder_repo = a10repo.VThunderRepository()
        self.listener_repo = repo.ListenerRepository()
        self.pool_repo = repo.PoolRepository()
        self.member_repo = repo.MemberRepository()
        self.cw = cw.A10ControllerWorker()
        self.svc_up_time = datetime.datetime.utcnow()

    def cleanup_slb_resources(self):
        cleanup_interval = CONF.a10_house_keeping.resource_cleanup_interval
        pending_lbs = self.loadbalancer_repo.get_pending_lbs_to_be_deleted(db_api.get_session(),
                                                                           cleanup_interval)
        for pending_lb in pending_lbs:
            self.cw.delete_load_balancer_with_housekeeping(pending_lb, True)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

MOCK_VTHUNDER_ID = 'mock-vthunder-1'
MOCK_HM_ID = 'mock-hm-cd61-49ed-8508-f22a12d770c5'
MOCK_POOL_ID = 'mock-pool-1'
MOCK_MEMBER_ID = 'mock-member-1'
MOCK_LOAD_BALANCER_ID = 'mock-lb-1'
MOCK_LISTENER_ID = 'mock-listener-1'
MOCK_FLAVOR_ID = 'mock-flavor-1'
MOCK_FLAVOR_PROF_ID = 'mock-flavor-prof-1'
MOCK_AMPHORA_ID = 'amphora_id'
MOCK_COMPUTE_ID = 'mock-compute-id'
MOCK_COOKIE_NAME = 'mock-cookie-1'
MOCK_CERT_FILENAME = 'mock-cert-filename-1'
MOCK_CERT_CONTENT = 'mock-cert-content-1'
MOCK_KEY_FILENAME = 'mock-key-filename-1'
MOCK_KEY_CONTENT = 'mock-key-content-1'
MOCK_KEY_PASS = 'mock-key-pass-1'
MOCK_TEMPLATE_NAME = 'mock-template-name-1'
MOCK_CERT_ACTION = 'import'
MOCK_CERT_TYPE = 'pem'

MOCK_VRID_FLOATING_IP_1 = '10.10.10.8'
MOCK_VRID_FLOATING_IP_2 = '10.10.10.9'
MOCK_VRID_PARTIAL_FLOATING_IP = '45'
MOCK_VRID_FULL_FLOATING_IP = '10.10.10.45'
MOCK_SUBNET_CIDR = '10.10.10.0/24'
MOCK_SUBNET_ID = 'mock-subnet-1'
MOCK_NETWORK_ID = 'mock-network-1'
MOCK_VRRP_PORT_ID = 'mock-vrrp-port-id'
MOCK_PORT_ID = 'mock-port-id'
MOCK_SET_ID = 'mock-set-id'
MOCK_FLEXPOOL_TOKEN = '0XD34DB33F'

SERVICE_GROUP_CONF_SECTION = 'service_group'
LISTENER_CONF_SECTION = 'listener'
HARDWARE_THUNDER_CONF_SECTION = 'hardware_thunder'
A10_GLOBAL_OPTS = 'a10_global'
MOCK_PARENT_PROJECT_ID = 'parent_project_id_with_length_32'
MOCK_CHILD_PART = 'child_part'
MOCK_CHILD_PROJECT_ID = '_child_project_id_with_length_32'
MOCK_CHILD_PARTITION = '_child_project'
MOCK_PROJECT_ID = 'mock-project-id'

A10_GLOBAL_CONF_SECTION = 'a10_global'
A10_HARDWARE_THUNDER_CONF_SECTION = 'hardware_thunder'
A10_CONTROLLER_WORKER_CONF_SECTION = 'a10_controller_worker'
LISTENER_CONF_SECTION = 'listener'
SERVICE_GROUP_CONF_SECTION = 'service_group'
SERVER_CONF_SECTION = 'server'
VIRTUAL_SERVER_CONFIG_SECTION = 'slb'
HEALTH_MONITOR_SECTION = 'health_monitor'
GLM_LICENSE_CONFIG_SECTION = 'glm_license'


A10_HOUSE_KEEPING = 'a10_house_keeping'

MOCK_SERVICE_GROUP_PROTOCOL = "HTTP"
MOCK_POOL_ID_2 = "mock-pool-2"
MOCK_SUBNET_ID_2 = "mock-subnet-2"

MOCK_L7POLICY_ID = "mock-l7-policy-1"
MOCK_L7RULE_ID = "mock-l7-rule-1"

MOCK_TLS_CERTIFICATE_ID = "mock_tls_certificate_id"
MOCK_TOPOLOGY_ACTIVE_STANDBY = "ACTIVE_STANDBY"
MOCK_TOPOLOGY_ROLE_MASTER = 'MASTER'
MOCK_TOPOLOGY_ROLE_BACKUP = 'BACKUP'

MOCK_OCTAVIA_HEALTH_MANAGER_CONTROLLER = "octavia_health_manager_controller"
MOCK_OCTAVIA_HEALTH_MONITOR = 'octavia_health_monitor'



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import cotyledon
from cotyledon import oslo_config_glue
import logging
import sys

from oslo_config import cfg
from oslo_reports import guru_meditation_report as gmr

from octavia import version

from a10_octavia.cmd import service as octavia_service
from a10_octavia.controller.queue import consumer

import multiprocessing

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
mp_mgr = multiprocessing.Manager()
ctx_map = mp_mgr.dict()
ctx_lock = mp_mgr.Lock()


def main():
    octavia_service.prepare_service(sys.argv)
    gmr.TextGuruMeditation.setup_autorun(version)
    sm = cotyledon.ServiceManager()
    sm.add(consumer.ConsumerService, workers=CONF.a10_controller_worker.workers,
           args=(CONF, ctx_map, ctx_lock))
    oslo_config_glue.setup(sm, CONF, reload_method="mutate")
    sm.run()



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from octavia.tests.unit import base


class BaseTaskTestCase(base.TestCase):

    def setUp(self):
        super(base.TestCase, self).setUp()
        patch('a10_octavia.controller.worker.tasks.decorators.axapi_client_decorator',
              lambda x: x).start()
        patch('a10_octavia.controller.worker.tasks.decorators.axapi_client_decorator_for_revert',
              lambda x: x).start()



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# ==============
# Argument Names
# ==============
LOADBALANCERS_LIST = 'loadbalancers_list'
VRID_LIST = 'vrid_list'
SET_THUNDER_UPDATE_AT = 'set-thunder-update-at'
SET_THUNDER_BACKUP_UPDATE_AT = 'set-thunder-backup-update-at'

# Member count with specific IP.
MEMBER_COUNT_IP = 'member_count_ip'
MEMBER_COUNT_IP_PORT_PROTOCOL = 'member_count_ip_port_protocol'
POOL_COUNT_IP = 'pool_count_ip'
WRITE_MEM_SHARED_PART = 'write_mem_shared_part'
WRITE_MEM_FOR_SHARED_PARTITION = 'write_memory_for_shared_partition'
WRITE_MEM_FOR_LOCAL_PARTITION = 'write_memory_for_local_partition'

MEMBER_LIST = 'member_list'
SUBNET_LIST = 'subnet_list'
MEMBERS = 'members'
POOLS = 'pools'

PARTITION_PROJECT_LIST = 'partition_project_list'
IFNUM_BACKUP = 'ifnum_backup'
IFNUM_MASTER = 'ifnum_master'
IPV6_ADDRESS_LIST = 'ifnum_address'
LISTENER_STATS = 'listener_stats'

# Octavia taskflow flow and task names (missing name in victoria octavia)
RELOADLOAD_BALANCER = 'octavia-reloadload-balancer'

VTHUNDER = 'vthunder'
STATUS = 'status'
ROLE = 'role'
BACKUP_VTHUNDER = 'backup_vthunder'
MASTER_VTHUNDER = 'master_vthunder'
VRRP_STATUS = 'vrrp_status'
MASTER_VRRP_STATUS = 'master_vrrp_status'
BACKUP_VRRP_STATUS = 'backup_vrrp_status'
NAT_POOL = 'nat_pool'
NAT_FLAVOR = 'nat_flavor'
SUBNET_PORT = 'subnet_port'
WRITE_MEM_SHARED = 'write_mem_shared'
WRITE_MEM_PRIVATE = 'write_mem_private'

FAILED = 'FAILED'
USED_SPARE = 'USED_SPARE'
SPARE_VTHUNDER = 'spare_vthunder'
FAILOVER_VTHUNDER = 'failover_vthunder'
OCTAVIA_HEALTH_MANAGER_CONTROLLER = 'octavia_health_manager_controller'
OCTAVIA_HEALTH_MONITOR = 'octavia_health_monitor'
VTHUNDER_CONFIG = 'vthunder_config'
DEVICE_CONFIG_DICT = 'device_config_dict'
DEVICE_KEY_PREFIX = '[dev]'
USE_DEVICE_FLAVOR = 'use_device_flavor'
VTHUNDER_ID = "vthunder_id"
DEVICE_ID = "device_id"
SET_ID = "set_id"
VRID = "vrid"
DEVICE_PRIORITY = 'device_priority'
FLOATING_IP = 'floating_ip'
FLOATING_IP_MASK = 'floating_ip_mask'
CERT_DATA = 'cert_data'
SP_OBJ_DICT = {
    'HTTP_COOKIE': "cookie_persistence",
    'APP_COOKIE': "cookie_persistence",
    'SOURCE_IP': "src_ip_persistence",
}
HTTP_TYPE = ['HTTP', 'HTTPS']
PERS_TYPE = ['cookie_persistence', 'src_ip_persistence']
NO_DEST_NAT_SUPPORTED_PROTOCOL = ['tcp', 'udp']
PORT = 'port'
LB_COUNT = 'lb_count'
LB_COUNT_SUBNET = 'lb_count_subnet'
LB_COUNT_FLAVOR = 'lb_count_flavor'
MEMBER_COUNT = 'member_count'
DELETE_VRID = 'delete_vrid'
LB_COUNT_THUNDER = 'lb_count_thunder'
MEMBER_COUNT_THUNDER = 'member_count_thunder'
LB_COUNT_THUNDER_PARTITION = 'lb_count_thunder_partition'

ID = 'id'
SECURITY_GROUPS = 'security_groups'
VIP_SEC_GROUP_PREFIX = 'lb-'
LB_RESOURCE = 'lb_resource'

SUBNET_ID = "subnet_id"
VLAN_ID = "vlan_id"
TAG_INTERFACE = "tag_interface"
VE_INTERFACE = "ve_interface"
DELETE_VLAN = "delete_vlan"
VLAN = "vlan"
FLAT = "flat"
SUPPORTED_NETWORK_TYPE = [FLAT, VLAN]
SSL_TEMPLATE = "ssl_template"

COMPUTE_BUSY = "compute_busy"
VTHUNDER_LIST = "vthunder_list"
NETWORK_LIST = "network_list"
ADDED_NETWORK = "added_network"

COMPUTE_PAUSED = "PAUSED"
L2DSR_FLAVOR = "l2dsr_flavor"
MASTER_AMPHORA_STATUS = "master_amphora_status"
BACKUP_AMPHORA_STATUS = "backup_amphora_status"

FLOW_TYPE = 'flow_type'
FLOW_TYPE_DELETE = 'delete'

# ACOS versions
ACOS_5_2_1_P2 = "5.2.1-P2"

# Proxy Protocol Object name
PROXY_POOL_COUNT = 'proxy_pool_count'
PROXY_PROTOCPL_TEMPLATE_NAME = 'octavia_proxy_protocol_v1'
PROXY_PROTOCPL_V2_TEMPLATE_NAME = 'octavia_proxy_protocol_v2'
PROXY_PROTOCPL_AFLEX_NAME = 'octavia_aflex_proxy_protocol_v1'
PROXY_PROTOCPL_V1 = "v1"
PROXY_PROTOCPL_V2 = "v2"

DUAL_STACK_MASK = 10000

# ============================
# Taskflow flow and task names
# ============================
GET_VTHUNDER_FOR_LB_SUBFLOW = 'octavia-get-vthunder-for-lb-subflow'
BACKUP_AMPHORA_PLUG = 'backup-amphora-plug'
GET_MASTER_VTHUNDER_INTERFACE = 'get-master-vthunder-intercae'
GET_BACKUP_VTHUNDER_INTERFACE = 'get-backup-vthunder-intercae'
MASTER_CONNECTIVITY_WAIT = 'master-connectivity-wait'
BACKUP_CONNECTIVITY_WAIT = 'backup-connectivity-wait'
ENABLE_MASTER_VTHUNDER_INTERFACE = 'enable-master-vthunder-interface'
ENABLE_BACKUP_VTHUNDER_INTERFACE = 'enable-backup-vthunder-interface'
ENABLE_VTHUNDER_INTERFACE = 'enable-vthunder-interface'
MASTER_ENABLE_INTERFACE = 'master-enable-interface'
BACKUP_ENABLE_INTERFACE = 'backup-enable-interface'
GET_IPV6_ADDRESS = 'get-ipv6-address'
GET_IPV6_ADDRESS_FOR_BACKUP = 'get-ipv6-address-for-backup'
MARK_VTHUNDER_MASTER_ACTIVE_IN_DB = 'mark-vthunder-master-active-in-db'
MARK_VTHUNDER_BACKUP_ACTIVE_IN_DB = 'mark-vthunder-backup-active-in-db'
MARK_VTHUNDER_MASTER_DELETED_IN_DB = 'mark-vthunder-master-deleted-in-db'
MARK_VTHUNDER_BACKUP_DELETED_IN_DB = 'mark-vthunder-backup-deleted-in-db'
GET_BACKUP_VTHUNDER_BY_LB = 'get-backup-vthunder-by-lb'
GET_VTHUNDER_FROM_LB = 'get-vthunder-from-lb'
CREATE_HEALTH_MONITOR_ON_VTHUNDER = 'create-health-monitor-on-vthunder'
CREATE_HEALTH_MONITOR_ON_SPARE = 'create-health-monitor-on-spare'
MARK_AMPHORA_READY_INDB = 'mark-amphora-ready-indb'
HANDLE_SESS_PERS = 'handle-session-persistence-delta-subflow'
GET_LOADBALANCER_FROM_DB = 'Get-Loadbalancer-from-db'
GET_BACKUP_LOADBALANCER_FROM_DB = 'get-backup-loadbalancer-from-db'
CONFIGURE_VRRP_FOR_MASTER_VTHUNDER = 'configure-vrrp-for-master-vthunder'
CONFIGURE_VRRP_FOR_BACKUP_VTHUNDER = 'configure-vrrp-for-backup-vthunder'
CONFIGURE_VRID_FOR_MASTER_VTHUNDER = 'configure-vrid-for-master-vthunder'
CONFIGURE_VRID_FOR_BACKUP_VTHUNDER = 'configure-vrid-for-backup-vthunder'
WAIT_FOR_MASTER_SYNC = 'wait-for-master-sync'
WAIT_FOR_BACKUP_SYNC = 'wait-for-backup-sync'
CONFIGURE_AVCS_SYNC_FOR_MASTER = 'configure-avcs-sync-for-master'
CONFIGURE_AVCS_SYNC_FOR_BACKUP = 'configure-avcs-sync-for-backup'
CONFIGURE_AVCS_FOR_FAILOVER = 'configure-avcs-for-failover'
CHECK_VRRP_STATUS = 'check-vrrp-status'
CHECK_VRRP_MASTER_STATUS = 'check-vrrp-master-status'
CHECK_VRRP_BACKUP_STATUS = 'check-vrrp-backup-status'
CONFIRM_VRRP_STATUS = 'confirm-vrrp-status'
ADD_VRRP_SET_ID_INDB = 'add-vrrp-set-id-db'
DELETE_VRRP_SET_ID_INDB = 'delete-vrrp-set-id-db'
GET_VRRP_SET_ID_INDB = 'get-vrrp-set-id-db'
ALLOCATE_VIP = 'allocate-vip'
UPDATE_VIP_AFTER_ALLOCATION = 'update-vip-after-allocation'
ALLOW_L2DSR = 'allow-l2dsr'
ALLOW_NO_SNAT = 'allow-no-snat'
AMPHORAE_POST_VIP_PLUG = 'amphorae-post-vip-plug'
AMPHORA_POST_NETWORK_UNPLUG = 'amphorae-post-network-plug'
AMPHORA_POST_NETWORK_UNPLUG_FOR_BACKUP_VTHUNDER = 'amphorae-post-network-plug-for-backup-vthunder'
AMP_POST_VIP_PLUG = 'amp-post-vip-plug'
AMPHORAE_POST_VIP_PLUG_FOR_MASTER = 'amphorae-post-vip-plug-for-master'
AMPHORAE_POST_VIP_PLUG_FOR_BACKUP = 'amphorae-post-vip-plug-for-backup'
VCS_RELOAD = 'vcs-reload'
GET_BACKUP_VTHUNDER = 'get-backup-vthunder'
GET_MASTER_VTHUNDER = 'get-master-vthunder'
GET_MASTER_IPV6_ADDRESS = 'get-master-ipv6-address'
GET_BACKUP_IPV6_ADDRESS = 'get-backup-ipv6-address'
CONNECTIVITY_WAIT_FOR_MASTER_VTHUNDER = 'connectivity-wait-for-master-vthunder'
CONNECTIVITY_WAIT_FOR_BACKUP_VTHUNDER = 'connectivity-wait-for-backup-vthunder'
GET_VTHUNDER_MASTER = 'get-vthunder-master'
WAIT_FOR_MASTER_VCS_RELOAD = 'wait-for-master-vcs-reload'
WAIT_FOR_BACKUP_VCS_RELOAD = 'wait-for-backup-vcs-reload'
VCS_SYNC_WAIT = "wait-for-vcs_ready"
GET_COMPUTE_FOR_PROJECT = 'get-compute-for-project'
VALIDATE_COMPUTE_FOR_PROJECT = 'validate-compute-for-project'
GET_SPARE_COMPUTE_FOR_PROJECT = 'get-spare-compute-for-project'
DELETE_STALE_SPARE_VTHUNDER = 'delete-stale-spare-vthunder'
CREATE_VTHUNDER_ENTRY = 'create-vthunder-entry'
UPDATE_ACOS_VERSION_IN_VTHUNDER_ENTRY = 'update-acos-version-in-vthunder-entry'
UPDATE_ACOS_VERSION_FOR_BACKUP_VTHUNDER = 'update-acos-version-for-backup-vthunder'
VTHUNDER_BY_LB = 'vthunder-by-loadbalancer'
GET_VTHUNDER_BY_LB = 'get-vthunder-by-lb'
VTHUNDER_CONNECTIVITY_WAIT = 'vthunder-connectivity-wait'
WAIT_FOR_VTHUNDER_CONNECTIVITY = 'wait-for-vthunder-connectivity'
CHANGE_PARTITION = 'change-partition'
CREATE_SSL_CERT_FLOW = 'create-ssl-cert-flow'
DELETE_SSL_CERT_FLOW = 'delete-ssl-cert-flow'
UPDATE_SSL_CERT_FLOW = 'update-ssl-cert-flow'
LISTENER_TYPE_DECIDER_FLOW = 'listener_type_decider_flow'
DELETE_LOADBALANCER_VRID_SUBFLOW = 'delete-loadbalancer-vrid-subflow'
REVOKE_ACTIVE_VTHUNDER_LICENSE = 'revoke-active-vthunder-license'
REVOKE_BACKUP_VTHUNDER_LICENSE = 'revoke-backup-vthunder-license'
HANDLE_VRID_LOADBALANCER_SUBFLOW = 'handle-vrid-loadbalancer-subflow'
CREATE_MEMBER_SNAT_POOL_SUBFLOW = 'create-member-snat-pool-subflow'
DELETE_MEMBER_VTHUNDER_INTERNAL_SUBFLOW = 'delete-member-vthunder-internal-subflow'
DELETE_MEMBER_VRID_SUBFLOW = 'delete-member-vrid-subflow'
DELETE_MEMBER_VRID_INTERNAL_SUBFLOW = 'delete-member-vrid-internal-subflow'
DELETE_HEALTH_MONITOR_VTHUNDER_SUBFLOW = 'delete-hm-vthunder-subflow'
DELETE_HEALTH_MONITOR_SUBFLOW_WITH_POOL_DELETE_FLOW = 'delete-health-monitor-subflow' \
                                                      '-with-pool-delete-flow'
DELETE_MEMBERS_SUBFLOW_WITH_POOL_DELETE_FLOW = 'delete-members-subflow-with-pool-delete-flow'
HANDLE_VRID_MEMBER_SUBFLOW = 'handle-vrid-member-subflow'
SPARE_VTHUNDER_CREATE = 'spare-vthunder-create'
ACTIVATE_GLM_LICENSE_SUBFLOW = 'activate-glm-license-subflow'
CONFIGURE_DNS_NAMESERVERS = 'configure-dns-nameservers'

ACTIVATE_FLEXPOOL_LICENSE = 'activate-flexpool-license'
CONFIGURE_PROXY_SERVER = 'configure-proxy-server'
SET_VTHUNDER_HOSTNAME = 'set-vthunder-hostname'
WRITE_MEMORY_THUNDER_FLOW = 'write-memory-thunder-flow'
RELOAD_CHECK_THUNDER_FLOW = 'reload-check-thunder-flow'
LB_TO_VTHUNDER_SUBFLOW = 'lb-to-vthunder-subflow'
GET_SPARE_AMPHORA_SBUFLOW = 'get-sapre_amphora-subflow'
GET_LB_RESOURCE = 'get-lb-resource'
GET_LB_RESOURCE_SUBNET = 'get-lb-resource-subnet'
GET_PROJECT_COUNT = 'get-child-parent-project-count'
GET_LB_COUNT_SUBNET = 'get-lb-count-by-subnet'
UPDATE_LISTENER_STATS_FLOW = 'update-listener-stats-flow'
GET_VTHUNDER_AMPHORA = 'get-vthunder-amphora'
COMPUTE_DELETE = 'compute-delete'
SET_VTHUNDER_TO_STANDBY = 'set-vthunder-to-standby'
GET_LBS_BY_THUNDER = 'get-loadbalancers-by-thunder'
MARK_LB_PENIND_UPDATE_IN_DB = 'mark-vthunder-pending-update-in-db'
MARK_LB_ACTIVE_IN_DB = 'mark-vthunder-active-in-db'
GET_VTHUNDER_NETWORK_LIST = 'get-vthunder-network-list'
GET_AMPHORA_FOR_FAILOVER = 'get-amphora-for-failover'
PLUG_NETWORK_BY_IDS = 'plug-network-by-ids'
PLUG_VIP_NETWORK_ON_SPARE = 'plusg-vip-network-on-spare'
POST_SPARE_PLUG_NETWORK = 'post-failover-plug-network'
GET_VCS_DEVICE_ID = 'get-vcs-device-id'
POST_FAILOVER_DB_UPDATE = 'post-failover-db-update'
MARK_LB_LIST_ERROR_ON_REVERT = 'mark-lb-list-error-on-revert'
GET_BACKUP_VTHUNDER_FROM_DB = 'get-backup-vthunder-from-db'

FULLY_POPULATED_LISTENER_CREATE = 'fully-populated-listener-create'
FULLY_POPULATED_ERROR_ON_REVERT = '-error-on-revert'
FULLY_POPULATED_GET_FLAVOR = '-get-flavor'
FULLY_POPULATED_NAT_CREATE = '-natpool-create'
FULLY_POPULATED_CREATE_LISTENER = '-create-listener'
FULLY_POPULATED_MARK_LISTENER_ACTIVE = '-mark-listener-active'
FULLY_POPULATED_GET_POOL_LISTENER = '-get-pool-listener'
FULLY_POPULATED_CREATE_POOL = '-create-pool'
FULLY_POPULATED_MARK_POOL_ACTIVE = '-mark-pool-active'
FULLY_POPULATED_VALID_SUBNET = '-validate-subnet'
FULLY_POPULATED_CREATE_MEMBER = '-create-member'
FULLY_POPULATED_MARK_MEMBER_ACTIVE = '-mark-member-active'
FULLY_POPULATED_CREATE_HM = '-create-hm'
FULLY_POPULATED_MARK_HM_ACTIVE = '-mark-hm-active'
FULLY_POPULATED_CREATE_L7POLICY = '-create-l7policy'
FULLY_POPULATED_MARK_L7POLICY_ACTIVE = '-mark-l7policy-active'
FULLY_POPULATED_CREATE_L7RULE = '-create-l7rule'
FULLY_POPULATED_MARK_L7RULE_ACTIVE = '-mark-l7rule-active'

# ======================
# Non-Taskflow Constants
# ======================

OCTAVIA_OWNER = 'Octavia'
TOPOLOGY_SPARE = 'SPARE'
READY = "READY"

IPV6_ADDRESS = 'ipv6-address'
IPV6_ADDRESS_CFG = 'ipv6-address-cfg'
IPV6_ADDRESS_PARTITION_CFG = 'ipv6-address-part-cfg'
IPV6_ADDRESS_PARTITION = 'ipv6-address-partition'
IP_ADDRESS = 'ip-address'
IP_ADDRESS_CFG = 'ip-address-cfg'
IP_ADDRESS_PARTITION_CFG = 'ip-address-part-cfg'
IP_ADDRESS_PARTITION = 'ip-address-partition'



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import re

from oslo_config import cfg

import acos_client.errors as acos_errors

from octavia.common import constants
from octavia_lib.common import constants as lib_consts

from a10_octavia.common import a10constants
from a10_octavia.common.data_models import Certificate
from a10_octavia.common import utils as a10_utils


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def get_cert_data(barbican_client, listener):
    cert_data = Certificate()
    cert_ref = listener.tls_certificate_id
    cert_containers = barbican_client.containers.list()
    for cert_container in cert_containers:
        if cert_container.container_ref == cert_ref:
            cert_data = Certificate(cert_filename=cert_container.certificate.name,
                                    key_filename=cert_container.private_key.name,
                                    cert_content=cert_container.certificate.payload,
                                    key_content=cert_container.private_key.payload,
                                    key_pass=cert_container.private_key_passphrase,
                                    template_name=listener.id)
    LOG.info("Secret container not found %s", cert_ref)
    return cert_data


def get_sess_pers_templates(pool):
    c_pers, s_pers, sp = None, None, None
    if pool and pool.session_persistence:
        sp = pool.session_persistence
        if hasattr(pool.session_persistence, 'to_dict'):
            sp = pool.session_persistence.to_dict()
        if sp['type'] == 'HTTP_COOKIE' or sp['type'] == 'APP_COOKIE':
            c_pers = pool.id
        elif sp['type'] == 'SOURCE_IP':
            s_pers = pool.id
    return c_pers, s_pers


def is_proxy_protocol_pool(pool):
    if pool.protocol == constants.PROTOCOL_PROXY or (
            pool.protocol == lib_consts.PROTOCOL_PROXYV2):
        return True
    return False


def proxy_protocol_use_aflex(listener, pool):
    if pool.protocol == constants.PROTOCOL_PROXY:
        if listener is not None and (listener.protocol == constants.PROTOCOL_TCP or
                                     listener.protocol == 'tcp'):
            use_aflex_proxy = CONF.service_group.use_aflex_proxy
            if use_aflex_proxy and use_aflex_proxy is True:
                return True
    return False


def get_tcp_proxy_template(listener, pool):
    tcp_proxy = None
    aflex = None
    if pool is None:
        return tcp_proxy, aflex
    if pool.provisioning_status != constants.PENDING_DELETE and (
            is_proxy_protocol_pool(pool) is True):
        if proxy_protocol_use_aflex(listener, pool) is True:
            aflex = a10constants.PROXY_PROTOCPL_AFLEX_NAME
        else:
            tcp_proxy = a10constants.PROXY_PROTOCPL_TEMPLATE_NAME
            if pool.protocol != constants.PROTOCOL_PROXY:
                tcp_proxy = a10constants.PROXY_PROTOCPL_V2_TEMPLATE_NAME
    return tcp_proxy, aflex


def get_proxy_aflex_list(curr, proxy_aflex, exclude_aflex):
    new_aflex_scripts = []
    if curr is not None and 'aflex-scripts' in curr['port']:
        aflexs = curr['port']['aflex-scripts']
        for aflex in aflexs:
            if exclude_aflex is None or aflex['aflex'] != exclude_aflex:
                new_aflex_scripts.append(aflex)
    if proxy_aflex is not None:
        new_aflex_scripts.append({"aflex": proxy_aflex})
    return new_aflex_scripts


def meta(lbaas_obj, key, default):
    if isinstance(lbaas_obj, dict):
        meta = lbaas_obj.get('a10_meta', '{}')
    elif hasattr(lbaas_obj, 'a10_meta'):
        meta = lbaas_obj.a10_meta
    else:
        return default
    try:
        meta_json = json.loads(meta)
    except Exception:
        return default
    return meta_json.get(key, default)


def shared_template_modifier(template_type, template_name, device_templates):
    resource_type = template_type.split('-', 1)[1]
    resource_list_key = "{0}-list".format(resource_type)
    if resource_list_key in device_templates:
        for device_template in device_templates['template'][resource_list_key]:
            device_template_name = device_template[resource_type].get("name")
            if template_name == device_template_name:
                break
            template_type = "{0}-shared".format(template_type)
    else:
        template_type = "{0}-shared".format(template_type)
    return template_type


def parse_name_expressions(name, name_expressions):
    flavor_data = {}
    if name and name_expressions:
        for expression in name_expressions:
            if 'regex' in expression:
                if re.search(expression['regex'], name):
                    flavor_data.update(expression['json'])
    return flavor_data


def dash_to_underscore(my_dict):
    if type(my_dict) is list:
        item_list = []
        for item in my_dict:
            item_list.append(dash_to_underscore(item))
        return item_list
    elif type(my_dict) is dict:
        item_dict = {}
        for k, v in my_dict.items():
            item_dict[k.replace('-', '_')] = dash_to_underscore(v)
        return item_dict
    else:
        return my_dict


def attribute_search(lb_resource, attr_name):
    """This helper method will recursively walk up the slb tree
    starting from the provided lb_resource and find an attribute
    with the provided name. Though objects like pool refrence a
    listener and loadbalancer, the following discrete
    search orders are used:

    1: member -> pool -> listener -> loadbalancer
    2: healthmonitor -> pool -> listener -> loadbalancer

    :param lb_resource: An slb data model
    :param obj_type: String name of an slb object
    (ie 'loadbalancer', 'pool')

    :return: Returns the requested attribute value or none
    """
    if hasattr(lb_resource, attr_name):
        return getattr(lb_resource, attr_name)
    elif hasattr(lb_resource, 'pool'):
        return attribute_search(lb_resource.pool, attr_name)
    elif hasattr(lb_resource, 'listener'):
        return attribute_search(lb_resource.listener, attr_name)
    elif hasattr(lb_resource, 'load_balancer'):
        return attribute_search(lb_resource.load_balancer, attr_name)
    return None


def get_member_server_name(axapi_client, member, raise_not_found=True):
    default_name = '{}_{}'.format(member.project_id[:5], member.ip_address.replace('.', '_'))
    server_name = default_name
    try:
        server_name = axapi_client.slb.server.get(server_name)
    except (acos_errors.NotFound):
        # Backwards compatability with a10-neutron-lbaas
        if CONF.a10_global.use_parent_partition:
            try:
                parent_project_id = a10_utils.get_parent_project(member.project_id)
                server_name = '_{}_{}_neutron'.format(parent_project_id[:5],
                                                      member.ip_address.replace('.', '_'))
                server_name = axapi_client.slb.server.get(server_name)
            except (acos_errors.NotFound):
                server_name = '_{}_{}_neutron'.format(member.project_id[:5],
                                                      member.ip_address.replace('.', '_'))
                try:
                    server_name = axapi_client.slb.server.get(server_name)
                except (acos_errors.NotFound) as e:
                    if raise_not_found:
                        raise e
                    return default_name
        else:
            try:
                server_name = axapi_client.slb.server.get('_{}_{}'.format(server_name, 'neutron'))
            except (acos_errors.NotFound) as e:
                if raise_not_found:
                    raise e
                return default_name
    return server_name['server']['name']


def acos_version_str2int(ver):
    if ver.isdigit():
        return int(ver)
    else:
        vnum = ""
        for index in range(len(ver)):
            if ver[index].isdigit():
                vnum = vnum + ver[index]
            else:
                break
        return int(vnum)


def acos_revision_parse(revision):
    rev = []
    revision_list = ['GR', 'P', 'SP']
    for tag in revision_list:
        tag_index = revision.find(tag)
        if tag_index >= 0:
            tag_len = len(tag)
            rev.append(acos_version_str2int(revision[(tag_index + tag_len):]))
        else:
            rev.append(0)

    return tuple(rev)


def acos_version(acos_version):
    major = acos_version_str2int(acos_version.split('.')[0])
    minor = acos_version_str2int(acos_version.split('.')[1])
    patch = acos_version_str2int(acos_version.split('.')[2])

    revision = ""
    prev = acos_version.find('-')
    if prev > 0:
        revision = acos_version[prev + 1:].upper()
    gr, p, sp = acos_revision_parse(revision)

    return (major, minor, patch, gr, p, sp)


def acos_version_cmp(ver1, ver2):
    vtup1 = acos_version(ver1)
    vtup2 = acos_version(ver2)
    for index in range(len(vtup1)):
        if vtup1[index] != vtup2[index]:
            return vtup1[index] - vtup2[index]
    return 0



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from taskflow.patterns import graph_flow
from taskflow.patterns import linear_flow

from octavia.common import constants
from octavia.controller.worker.v1.tasks import database_tasks
from octavia.controller.worker.v1.tasks import lifecycle_tasks
from octavia.controller.worker.v1.tasks import model_tasks

from a10_octavia.common import a10constants
from a10_octavia.controller.worker.flows import a10_health_monitor_flows
from a10_octavia.controller.worker.flows import a10_member_flows
from a10_octavia.controller.worker.tasks import a10_database_tasks
from a10_octavia.controller.worker.tasks import a10_network_tasks
from a10_octavia.controller.worker.tasks import persist_tasks
from a10_octavia.controller.worker.tasks import service_group_tasks
from a10_octavia.controller.worker.tasks import virtual_port_tasks
from a10_octavia.controller.worker.tasks import vthunder_tasks

from oslo_config import cfg
CONF = cfg.CONF


class PoolFlows(object):

    def __init__(self):
        self.hm_flow = a10_health_monitor_flows.HealthMonitorFlows()
        self.member_flow = a10_member_flows.MemberFlows()

    def get_create_pool_flow(self, topology):
        """Create a flow to create a pool

        :returns: The flow for creating a pool
        """

        create_pool_flow = linear_flow.Flow(constants.CREATE_POOL_FLOW)
        create_pool_flow.add(lifecycle_tasks.PoolToErrorOnRevertTask(
            requires=[constants.POOL,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        create_pool_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        create_pool_flow.add(database_tasks.MarkPoolPendingCreateInDB(
            requires=constants.POOL))
        create_pool_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            create_pool_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        create_pool_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.POOL},
            provides=constants.FLAVOR))
        create_pool = service_group_tasks.PoolCreate(
            requires=[constants.POOL, a10constants.VTHUNDER, constants.FLAVOR, constants.LISTENER],
            provides=constants.POOL)
        create_pool_flow.add(*self._get_sess_pers_subflow(create_pool))
        create_pool_flow.add(virtual_port_tasks.ListenerUpdateForPool(
            requires=[constants.LOADBALANCER, constants.LISTENER,
                      constants.POOL, a10constants.VTHUNDER]))
        create_pool_flow.add(database_tasks.MarkPoolActiveInDB(
            requires=constants.POOL))
        create_pool_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        create_pool_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        create_pool_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))

        return create_pool_flow

    def get_fully_populated_create_pool_flow(self, topology, pool, vthunder_conf=None,
                                             device_dict=None, vthunder_flow=False):
        """Create pool for fully populated creation"""

        sf_name = constants.CREATE_POOL_FLOW + '_' + pool.id
        create_pool_flow = linear_flow.Flow(sf_name)
        create_pool_flow.add(service_group_tasks.PoolToErrorOnRevertTask(
            name=sf_name + a10constants.FULLY_POPULATED_ERROR_ON_REVERT,
            requires=constants.POOL,
            inject={constants.POOL: pool}))

        if pool.health_monitor:
            create_pool_flow.add(self.hm_flow.get_fully_populated_create_health_monitor_flow(
                topology, pool.health_monitor))

        create_pool_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            name=sf_name + a10constants.GET_VTHUNDER_BY_LB,
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            create_pool_flow.add(vthunder_tasks.GetMasterVThunder(
                name=sf_name + '_' + a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        create_pool_flow.add(a10_database_tasks.GetFlavorData(
            name=sf_name + a10constants.FULLY_POPULATED_GET_FLAVOR,
            inject={constants.POOL: pool},
            rebind={a10constants.LB_RESOURCE: constants.POOL},
            provides=constants.FLAVOR))
        create_pool_flow.add(a10_database_tasks.GetPoolListener(
            name=sf_name + a10constants.FULLY_POPULATED_GET_POOL_LISTENER,
            requires=constants.POOL,
            inject={constants.POOL: pool},
            provides=constants.LISTENER))
        create_pool = service_group_tasks.PoolCreate(
            name=sf_name + a10constants.FULLY_POPULATED_CREATE_POOL,
            requires=[constants.POOL, a10constants.VTHUNDER, constants.FLAVOR, constants.LISTENER],
            inject={constants.POOL: pool},
            provides=constants.POOL)
        create_pool_flow.add(*self._get_sess_pers_subflow(create_pool, sf_name))
        create_pool_flow.add(database_tasks.MarkPoolActiveInDB(
            name=sf_name + a10constants.FULLY_POPULATED_MARK_POOL_ACTIVE,
            requires=constants.POOL,
            inject={constants.POOL: pool}))

        for member in pool.members:
            if vthunder_flow:
                create_pool_flow.add(
                    self.member_flow.get_vthunder_fully_populated_create_member_flow(
                        topology, member))
            else:
                create_pool_flow.add(self.member_flow.get_rack_fully_populated_create_member_flow(
                    vthunder_conf, device_dict, member))

        return create_pool_flow

    def get_delete_pool_flow(self, members, health_mon, store, topology):
        """Create a flow to delete a pool

        :returns: The flow for deleting a pool
        """
        delete_pool_flow = linear_flow.Flow(constants.DELETE_POOL_FLOW)
        delete_pool_flow.add(lifecycle_tasks.PoolToErrorOnRevertTask(
            requires=[constants.POOL,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        delete_pool_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        delete_pool_flow.add(database_tasks.MarkPoolPendingDeleteInDB(
            requires=constants.POOL))
        delete_pool_flow.add(database_tasks.CountPoolChildrenForQuota(
            requires=constants.POOL, provides=constants.POOL_CHILD_COUNT))
        delete_pool_flow.add(model_tasks.DeleteModelObject(
            rebind={constants.OBJECT: constants.POOL}))
        # Get VThunder details from database
        delete_pool_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            delete_pool_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        delete_pool_flow.add(virtual_port_tasks.ListenerUpdateForPool(
            requires=[constants.LOADBALANCER, constants.LISTENER,
                      constants.POOL, a10constants.VTHUNDER]))
        delete_pool_flow.add(persist_tasks.DeleteSessionPersistence(
            requires=[a10constants.VTHUNDER, constants.POOL]))
        delete_pool_flow.add(a10_database_tasks.GetProxyProtocolPoolCount(
            requires=[constants.LISTENER, constants.POOL],
            provides=a10constants.PROXY_POOL_COUNT))
        # Delete pool children
        delete_pool_flow.add(self._get_delete_member_vthunder_subflow(members, store))
        delete_pool_flow.add(self._get_delete_health_monitor_vthunder_subflow(health_mon))
        delete_pool_flow.add(service_group_tasks.PoolDelete(
            requires=[constants.POOL, a10constants.VTHUNDER, constants.LISTENER,
                      a10constants.PROXY_POOL_COUNT]))
        delete_pool_flow.add(database_tasks.DeletePoolInDB(
            requires=constants.POOL))
        # Interface delete.
        delete_pool_flow.add(a10_database_tasks.GetLoadBalancerListByProjectID(
            requires=a10constants.VTHUNDER,
            provides=a10constants.LOADBALANCERS_LIST))
        delete_pool_flow.add(a10_database_tasks.GetMemberListByProjectID(
            requires=a10constants.VTHUNDER,
            provides=a10constants.MEMBER_LIST))
        delete_pool_flow.add(a10_network_tasks.CalculateDelta(
            requires=(constants.LOADBALANCER, a10constants.LOADBALANCERS_LIST,
                      a10constants.MEMBER_LIST),
            provides=constants.DELTAS))
        delete_pool_flow.add(a10_network_tasks.HandleNetworkDeltas(
            requires=constants.DELTAS, provides=constants.ADDED_PORTS))
        delete_pool_flow.add(
            vthunder_tasks.AmphoraePostNetworkUnplug(
                requires=(
                    constants.LOADBALANCER,
                    constants.ADDED_PORTS,
                    a10constants.VTHUNDER)))
        delete_pool_flow.add(database_tasks.GetAmphoraeFromLoadbalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.AMPHORA))
        delete_pool_flow.add(vthunder_tasks.VThunderComputeConnectivityWait(
            requires=(a10constants.VTHUNDER, constants.AMPHORA)))
        delete_pool_flow.add(database_tasks.DecrementPoolQuota(
            requires=[constants.POOL, constants.POOL_CHILD_COUNT]))
        delete_pool_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        delete_pool_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        delete_pool_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))

        return delete_pool_flow

    def _get_delete_health_monitor_vthunder_subflow(self, health_mon, cascade=False):
        delete_hm_vthunder_subflow = linear_flow.Flow(
            a10constants.DELETE_HEALTH_MONITOR_SUBFLOW_WITH_POOL_DELETE_FLOW)
        if health_mon:
            if cascade:
                delete_hm_vthunder_subflow.add(
                    self.hm_flow.get_delete_health_monitor_vthunder_subflow(health_mon))
            else:
                delete_hm_vthunder_subflow.add(
                    self.hm_flow.get_delete_health_monitor_vthunder_subflow())
        return delete_hm_vthunder_subflow

    def _get_delete_member_vthunder_subflow(self, members, store, pool=constants.POOL):
        delete_member_vthunder_subflow = linear_flow.Flow(
            a10constants.DELETE_MEMBERS_SUBFLOW_WITH_POOL_DELETE_FLOW)
        member_store = {}
        for member in members:
            member_store[member.id] = member
            delete_member_vthunder_subflow.add(
                self.member_flow.get_delete_member_vthunder_internal_subflow(member.id, pool))
        if members and CONF.a10_global.handle_vrid:
            delete_member_vthunder_subflow.add(
                self.member_flow.get_delete_member_vrid_internal_subflow(pool, members))
        store.update(member_store)
        return delete_member_vthunder_subflow

    def get_update_pool_flow(self, topology):
        """Create a flow to update a pool

        :returns: The flow for updating a pool
        """
        update_pool_flow = linear_flow.Flow(constants.UPDATE_POOL_FLOW)
        update_pool_flow.add(lifecycle_tasks.PoolToErrorOnRevertTask(
            requires=[constants.POOL,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        update_pool_flow.add(vthunder_tasks.VthunderInstanceBusy(
            requires=a10constants.COMPUTE_BUSY))

        update_pool_flow.add(database_tasks.MarkPoolPendingUpdateInDB(
            requires=constants.POOL))
        update_pool_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            update_pool_flow.add(vthunder_tasks.GetMasterVThunder(
                name=a10constants.GET_MASTER_VTHUNDER,
                requires=a10constants.VTHUNDER,
                provides=a10constants.VTHUNDER))
        update_pool_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.POOL},
            provides=constants.FLAVOR))
        update_pool = service_group_tasks.PoolUpdate(
            requires=[constants.POOL, a10constants.VTHUNDER,
                      constants.UPDATE_DICT, constants.FLAVOR],
            provides=constants.POOL)
        update_pool_flow.add(*self._get_sess_pers_subflow(update_pool))
        update_pool_flow.add(virtual_port_tasks.ListenerUpdateForPool(
            requires=[constants.LOADBALANCER, constants.LISTENER,
                      constants.POOL, a10constants.VTHUNDER]))
        update_pool_flow.add(database_tasks.UpdatePoolInDB(
            requires=[constants.POOL, constants.UPDATE_DICT]))
        update_pool_flow.add(database_tasks.MarkPoolActiveInDB(
            requires=constants.POOL))
        update_pool_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        update_pool_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        update_pool_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))

        return update_pool_flow

    def _get_sess_pers_subflow(self, pool_task, prefix=""):
        get_pool_create_with_sess_pers = graph_flow.Flow(a10constants.HANDLE_SESS_PERS)
        sess_pers = persist_tasks.HandleSessionPersistenceDelta(
            name=prefix + 'hanlde_session_persist_dela',
            requires=[a10constants.VTHUNDER, constants.POOL])
        get_pool_create_with_sess_pers.add(pool_task, sess_pers)
        get_pool_create_with_sess_pers.link(pool_task, sess_pers,
                                            decider=self._is_sess_pers_decider)
        return get_pool_create_with_sess_pers

    def _is_sess_pers_decider(self, history):
        """Decides if the pool has session persistence
        :returns: True if if pool has session persistence
        """
        return history[history.keys()[0]].session_persistence is not None

    def get_cascade_delete_pool_internal_flow(
            self, pool_name, members, pool_listener_name, health_mon):
        """Create a flow to delete a pool, etc.
        :returns: The flow for deleting a pool
        """
        store = {}
        delete_pool_flow = linear_flow.Flow(constants.DELETE_POOL_FLOW)
        # health monitor should cascade
        # members should cascade
        delete_pool_flow.add(database_tasks.MarkPoolPendingDeleteInDB(
            name='mark_pool_pending_delete_in_db_' + pool_name,
            requires=constants.POOL,
            rebind={constants.POOL: pool_name}))
        delete_pool_flow.add(database_tasks.CountPoolChildrenForQuota(
            name='count_pool_children_for_quota_' + pool_name,
            requires=constants.POOL,
            provides=constants.POOL_CHILD_COUNT,
            rebind={constants.POOL: pool_name}))
        delete_pool_flow.add(virtual_port_tasks.ListenerUpdateForPool(
            name='listener_update_for_pool_' + pool_name,
            requires=[constants.LOADBALANCER, constants.LISTENER, constants.POOL,
                      a10constants.VTHUNDER, a10constants.FLOW_TYPE],
            inject={a10constants.FLOW_TYPE: a10constants.FLOW_TYPE_DELETE},
            rebind={constants.LISTENER: pool_listener_name,
                    constants.POOL: pool_name}))
        delete_pool_flow.add(persist_tasks.DeleteSessionPersistence(
            name='delete_session_persistence_' + pool_name,
            requires=[a10constants.VTHUNDER, constants.POOL],
            rebind={constants.POOL: pool_name}))
        delete_pool_flow.add(model_tasks.DeleteModelObject(
            name='delete_model_object_' + pool_name,
            rebind={constants.OBJECT: pool_name}))
        delete_pool_flow.add(a10_network_tasks.GetPoolsOnThunder(
            name='get_pools_on_thunder_' + pool_name,
            requires=[a10constants.VTHUNDER, a10constants.USE_DEVICE_FLAVOR],
            provides=a10constants.POOLS))
        delete_pool_flow.add(a10_database_tasks.GetProxyProtocolPoolCount(
            name='get_proxy_pool_count_' + pool_name,
            requires=[constants.LISTENER, constants.POOL],
            rebind={constants.POOL: pool_name, constants.LISTENER: pool_listener_name},
            provides=a10constants.PROXY_POOL_COUNT))
        # Delete pool children
        delete_pool_flow.add(self._get_delete_member_vthunder_subflow(members, store, pool_name))
        delete_pool_flow.add(self._get_delete_health_monitor_vthunder_subflow(health_mon, True))
        delete_pool_flow.add(service_group_tasks.PoolDelete(
            name='pool_delete_' + pool_name,
            requires=[constants.POOL, a10constants.VTHUNDER, constants.LISTENER,
                      a10constants.PROXY_POOL_COUNT],
            rebind={constants.POOL: pool_name, constants.LISTENER: pool_listener_name}))
        delete_pool_flow.add(database_tasks.DeletePoolInDB(
            name='delete_pool_in_db_' + pool_name,
            requires=constants.POOL,
            rebind={constants.POOL: pool_name}))
        delete_pool_flow.add(database_tasks.DecrementPoolQuota(
            name='decrement_pool_quota_' + pool_name,
            requires=[constants.POOL, constants.POOL_CHILD_COUNT],
            rebind={constants.POOL: pool_name}))

        return (delete_pool_flow, store)

    def get_delete_pool_rack_flow(self, members, health_mon, store):
        """Create a flow to delete a pool rack

        :returns: The flow for deleting a pool
        """
        delete_pool_flow = linear_flow.Flow(constants.DELETE_POOL_FLOW)
        delete_pool_flow.add(lifecycle_tasks.PoolToErrorOnRevertTask(
            requires=[constants.POOL,
                      constants.LISTENERS,
                      constants.LOADBALANCER]))
        delete_pool_flow.add(database_tasks.MarkPoolPendingDeleteInDB(
            requires=constants.POOL))
        delete_pool_flow.add(database_tasks.CountPoolChildrenForQuota(
            requires=constants.POOL, provides=constants.POOL_CHILD_COUNT))
        delete_pool_flow.add(model_tasks.DeleteModelObject(
            rebind={constants.OBJECT: constants.POOL}))
        # Get VThunder details from database
        delete_pool_flow.add(a10_database_tasks.GetVThunderByLoadBalancer(
            requires=constants.LOADBALANCER,
            provides=a10constants.VTHUNDER))
        delete_pool_flow.add(virtual_port_tasks.ListenerUpdateForPool(
            requires=[constants.LOADBALANCER, constants.LISTENER,
                      constants.POOL, a10constants.VTHUNDER]))
        delete_pool_flow.add(persist_tasks.DeleteSessionPersistence(
            requires=[a10constants.VTHUNDER, constants.POOL]))
        delete_pool_flow.add(vthunder_tasks.SetupDeviceNetworkMap(
            requires=a10constants.VTHUNDER,
            provides=a10constants.VTHUNDER))

        # Device Flavor
        delete_pool_flow.add(a10_database_tasks.GetFlavorData(
            rebind={a10constants.LB_RESOURCE: constants.LOADBALANCER},
            provides=constants.FLAVOR))
        delete_pool_flow.add(vthunder_tasks.GetVthunderConfByFlavor(
            requires=(constants.LOADBALANCER, a10constants.VTHUNDER_CONFIG,
                      a10constants.DEVICE_CONFIG_DICT),
            rebind={constants.FLAVOR_DATA: constants.FLAVOR},
            provides=(a10constants.VTHUNDER_CONFIG, a10constants.USE_DEVICE_FLAVOR)))
        delete_pool_flow.add(a10_network_tasks.GetPoolsOnThunder(
            requires=[a10constants.VTHUNDER, a10constants.USE_DEVICE_FLAVOR],
            provides=a10constants.POOLS))
        delete_pool_flow.add(a10_database_tasks.GetProxyProtocolPoolCount(
            requires=[constants.LISTENER, constants.POOL],
            provides=a10constants.PROXY_POOL_COUNT))

        # Delete pool children
        delete_pool_flow.add(self._get_delete_member_vthunder_subflow(members, store))
        delete_pool_flow.add(self._get_delete_health_monitor_vthunder_subflow(health_mon))
        delete_pool_flow.add(service_group_tasks.PoolDelete(
            requires=[constants.POOL, a10constants.VTHUNDER, constants.LISTENER,
                      a10constants.PROXY_POOL_COUNT]))
        delete_pool_flow.add(database_tasks.DeletePoolInDB(
            requires=constants.POOL))
        delete_pool_flow.add(database_tasks.DecrementPoolQuota(
            requires=[constants.POOL, constants.POOL_CHILD_COUNT]))
        delete_pool_flow.add(database_tasks.MarkLBAndListenersActiveInDB(
            requires=[constants.LOADBALANCER, constants.LISTENERS]))
        delete_pool_flow.add(vthunder_tasks.WriteMemory(
            requires=a10constants.VTHUNDER))
        delete_pool_flow.add(a10_database_tasks.SetThunderUpdatedAt(
            requires=a10constants.VTHUNDER))

        return delete_pool_flow



# Copyright 2019 A10 Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""A10 Octavia Helper Module

"""
import acos_client
import ipaddress
import json
import netaddr
import socket
import struct

from ipaddress import ip_address
from ipaddress import IPv4Address
from ipaddress import IPv6Address
from oslo_config import cfg
from oslo_log import log as logging

from keystoneauth1.exceptions import http as keystone_exception
from keystoneclient.v3 import client as keystone_client
from octavia.common import keystone
from octavia.db import api as db_apis
from octavia.db import repositories as repo
from stevedore import driver as stevedore_driver

from a10_octavia.common import a10constants
from a10_octavia.common import data_models
from a10_octavia.common import exceptions

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def validate_ipv4(address):
    """Validate for IP4 address format"""
    if not netaddr.valid_ipv4(address, netaddr.core.INET_PTON):
        raise cfg.ConfigFileValueError(
            'Invalid IPAddress value given in configuration: {0}'.format(address))


def validate_ipv6(IP: str):
    try:
        ipaddress.ip_address(IP)
        LOG.debug("Valid IP address")
    except ValueError:
        raise cfg.ConfigFileValueError(
            'Invalid IPAddress value given in configuration: {0}'.format(IP))


def validate_partial_ipv6(IP: str):
    try:
        ipaddress.ip_address(IP)
        LOG.debug("Valid IP address")
        return True
    except ValueError:
        return False


def validate_partial_ipv4(address):
    """Validate the partial IPv4 suffix provided"""
    partial_ip = address.lstrip('.')
    octets = partial_ip.split('.')
    for idx in range(4 - len(octets)):
        octets.insert(0, '0')
    if not netaddr.valid_ipv4('.'.join(octets), netaddr.core.INET_PTON):
        raise cfg.ConfigFileValueError('Invalid partial IPAddress value given'
                                       ' in configuration: {0}'.format(address))
    return address


def validate_partition(hardware_device):
    partition_name = hardware_device.get('partition_name', '')
    if len(partition_name) > 14:
        raise ValueError("Supplied partition value '%s' exceeds maximum length 14" %
                         (partition_name))


def validate_params(hardware_info):
    """Check for all the required parameters for hardware configurations."""
    if all(k in hardware_info for k in ('ip_address', 'username', 'password', 'device_name')):
        if all(hardware_info[x] is not None for x in ('ip_address',
                                                      'username', 'password', 'device_name')):
            validate_ipv4(hardware_info['ip_address'])
            validate_partition(hardware_info)
            return hardware_info
    raise cfg.ConfigFileValueError('Please check your configuration. The params `project_id`, '
                                   '`ip_address`, `username`, `password` and `device_name` '
                                   'under [hardware_thunder] section cannot be None ')


def check_duplicate_entries(hardware_dict):
    hardware_count_dict = {}
    for hardware_device in hardware_dict.values():
        if (hardware_device.hierarchical_multitenancy != "enable" and
                hardware_device.device_name_as_key is not True):
            candidate = '{}:{}'.format(hardware_device.ip_address, hardware_device.partition_name)
            hardware_count_dict[candidate] = hardware_count_dict.get(candidate, 0) + 1
    return [k for k, v in hardware_count_dict.items() if v > 1]


def convert_to_hardware_thunder_conf(hardware_list):
    """Validate for all vthunder nouns for hardware devices configurations."""
    hardware_dict = {}
    for hardware_device in hardware_list:
        hardware_device = validate_params(hardware_device)
        hardware_device['undercloud'] = True
        if hardware_device.get('interface_vlan_map'):
            hardware_device['device_network_map'] = validate_interface_vlan_map(hardware_device)
            del hardware_device['interface_vlan_map']

        # Add dict entries with project_id as key
        if 'project_id' in hardware_device:
            project_id = hardware_device['project_id']
            if hardware_dict.get(project_id):
                raise cfg.ConfigFileValueError('Supplied duplicate project_id {} '
                                               ' in [hardware_thunder] section'.format(project_id))
            hierarchical_mt = hardware_device.get('hierarchical_multitenancy')
            if hierarchical_mt == "enable":
                hardware_device["partition_name"] = project_id[0:14]
            if hierarchical_mt and hierarchical_mt not in ('enable', 'disable'):
                raise cfg.ConfigFileValueError('Option `hierarchical_multitenancy` specified '
                                               'under project id {} only accepts "enable" and '
                                               '"disable"'.format(project_id))
            vthunder_conf = data_models.HardwareThunder(**hardware_device)
            hardware_dict[project_id] = vthunder_conf

        # Add dict entries with dev_name as key
        device_name = hardware_device['device_name']
        if hardware_dict.get(a10constants.DEVICE_KEY_PREFIX + device_name):
            raise cfg.ConfigFileValueError('Supplied duplicate device_name {} '
                                           ' in [hardware_thunder] section'.format(device_name))
        vthunder_conf = data_models.HardwareThunder(**hardware_device)
        vthunder_conf.device_name_as_key = True
        hardware_dict[(a10constants.DEVICE_KEY_PREFIX + device_name)] = vthunder_conf

        duplicates = check_duplicate_entries(hardware_dict)
        if duplicates:
            raise cfg.ConfigFileValueError('Duplicates found for the following '
                                           '\'ip_address:partition_name\' entries: {}'
                                           .format(list(duplicates)))
    return hardware_dict


def get_vip_security_group_name(port_id):
    if port_id:
        return a10constants.VIP_SEC_GROUP_PREFIX + port_id
    return None


def get_parent_project_list():
    parent_project_list = []
    for project_id in CONF.hardware_thunder.devices:
        parent_project_id = get_parent_project(project_id)
        if parent_project_id != 'default':
            parent_project_list.append(parent_project_id)
    return parent_project_list


def get_parent_project(project_id):
    key_session = keystone.KeystoneSession().get_session()
    key_client = keystone_client.Client(session=key_session)
    try:
        project = key_client.projects.get(project_id)
        return project.parent_id
    except keystone_exception.NotFound:
        return None


def get_axapi_client(vthunder):
    api_ver = acos_client.AXAPI_21 if vthunder.axapi_version == 21 else acos_client.AXAPI_30
    axapi_client = acos_client.Client(vthunder.ip_address, api_ver,
                                      vthunder.username, vthunder.password,
                                      timeout=CONF.vthunder.default_axapi_timeout)
    return axapi_client


def get_net_info_from_cidr(cidr, ip_version):
    if ip_version == 4:
        subnet_ip, mask = cidr.split('/')
        avail_hosts = (1 << 32 - int(mask))
        netmask = socket.inet_ntoa(struct.pack('>I', (1 << 32) - avail_hosts))
        return subnet_ip, netmask
    else:
        ipv6_ip, ipv6_mask = ipaddress.IPv6Interface(cidr).with_netmask.split('/')
        return ipv6_ip, ipv6_mask


def check_ip_in_subnet_range(ip, subnet, netmask, ip_version, subnet_cidr):
    if ip_version == 4:
        if ip is None or subnet is None or netmask is None:
            return False
        int_ip = struct.unpack('>L', socket.inet_aton(ip))[0]
        int_subnet = struct.unpack('>L', socket.inet_aton(subnet))[0]
        int_netmask = struct.unpack('>L', socket.inet_aton(netmask))[0]
        return int_ip & int_netmask == int_subnet
    else:
        if ip is None or subnet_cidr is None:
            return False
        return ipaddress.IPv6Address(ip) in ipaddress.IPv6Network(subnet_cidr)


def merge_host_and_network_ip(cidr, host_ip):
    network_ip, mask = cidr.split('/')
    host_bits = struct.pack('>L', ((1 << (32 - int(mask))) - 1))
    int_host_bits = struct.unpack('>L', host_bits)[0]
    int_host_ip = struct.unpack('>L', socket.inet_aton(host_ip))[0]
    int_net_ip = struct.unpack('>L', socket.inet_aton(network_ip))[0]
    full_ip_packed = struct.pack('>L', int_net_ip | (int_host_ip & int_host_bits))
    return socket.inet_ntoa(full_ip_packed)


def get_network_driver():
    CONF.import_group('a10_controller_worker', 'a10_octavia.common.config_options')
    network_driver = stevedore_driver.DriverManager(
        namespace='octavia.network.drivers',
        name=CONF.a10_controller_worker.network_driver,
        invoke_on_load=True
    ).driver
    return network_driver


def get_patched_ip_address(ip, cidr, ip_version):
    net_ip, netmask = get_net_info_from_cidr(cidr, ip_version)
    if ip_version == 6:
        ip = ip.replace('.', "")
        is_valid_ip = validate_partial_ipv6(ip)
        if not is_valid_ip:
            ip = str(ipaddress.IPv6Address(net_ip)) + str(ip.replace(':', ""))
        if check_ip_in_subnet_range(ip, net_ip, netmask, ip_version, cidr):
            return ip
        else:
            raise exceptions.VRIDIPNotInSubentRangeError(ip, cidr)
    else:
        octets = ip.lstrip('.').split('.')

        if len(octets) == 4:
            validate_ipv4(ip)
            if check_ip_in_subnet_range(ip, net_ip, netmask, 4, None):
                return ip
            else:
                raise exceptions.VRIDIPNotInSubentRangeError(ip, cidr)

        for i in range(4 - len(octets)):
            octets.insert(0, '0')
        host_ip = '.'.join(octets)

        try:
            int_host_ip = struct.unpack('>L', socket.inet_aton(host_ip))[0]
        except socket.error:
            raise exceptions.VRIDIPNotInSubentRangeError(host_ip, cidr)

        int_net_ip = struct.unpack('>L', socket.inet_aton(net_ip))[0]
        int_netmask = struct.unpack('>L', socket.inet_aton(netmask))[0]

        # Create a set of test bits to find partial netmask octet
        # ie 1.1.1.1 & 255.255.254.0 -> 1.1.0.0
        test_bits = (1 << 24 | 1 << 16 | 1 << 8 | 1)
        test_bits = (int_netmask & test_bits)

        # Then shift 8 bits (1 -> 256) and subtract by 1 in each octet to get 255
        test_bits = (test_bits << 8) - test_bits

        # Use truncated mask to fill in any missing octets of partial IP
        canidate_ip = (test_bits & int_net_ip) | int_host_ip

        if (canidate_ip & int_netmask) != (int_net_ip & int_netmask):
            raise exceptions.VRIDIPNotInSubentRangeError(host_ip, cidr)

        return socket.inet_ntoa(struct.pack('>L', canidate_ip))


def get_vrid_floating_ip_for_project(project_id):
    vrid_fp = None
    if CONF.hardware_thunder.devices:
        device_info = CONF.hardware_thunder.devices.get(project_id)
        if device_info:
            vrid_fp = device_info.vrid_floating_ip
    return CONF.a10_global.vrid_floating_ip if not vrid_fp else vrid_fp


def validate_vcs_device_info(device_network_map):
    num_devices = len(device_network_map)
    if num_devices > 2:
        raise exceptions.VcsDevicesNumberExceedsConfigError(num_devices)
    for device_obj in device_network_map:
        device_id = device_obj.vcs_device_id
        if ((num_devices > 1 and not device_id) or
            (device_id is not None and (not isinstance(device_id, int) or
                                        (device_id < 1 or device_id > 2)))):
            raise exceptions.InvalidVcsDeviceIdConfigError(device_id)
        if num_devices > 1:
            if not device_obj.mgmt_ip_address:
                raise exceptions.MissingMgmtIpConfigError(device_id)
            else:
                validate_ipv4(device_obj.mgmt_ip_address)


def convert_interface_to_data_model(interface_obj):
    vlan_map_list = interface_obj.get('vlan_map')
    interface_num = interface_obj.get('interface_num')
    interface_dm = data_models.Interface()
    if not interface_num:
        raise exceptions.MissingInterfaceNumConfigError()
    if not type(interface_num) == int:
        raise exceptions.InvalidInterfaceNumberConfigError(interface_num)
    if not vlan_map_list:
        LOG.warning("Empty vlan map provided in configuration file")
    for vlan_map in vlan_map_list:
        vlan_id = vlan_map.get('vlan_id')
        if not vlan_id:
            raise exceptions.MissingVlanIDConfigError(interface_num)
        if not type(vlan_id) == int:
            raise exceptions.InvalidVlanIdConfigError(vlan_id)
        if vlan_id in interface_dm.tags:
            raise exceptions.DuplicateVlanTagsConfigError(interface_num, vlan_id)
        if vlan_map.get('use_dhcp') and vlan_map.get('use_dhcp') not in ("True", "False"):
            raise exceptions.InvalidUseDhcpConfigError(vlan_map.get('use_dhcp'))
        if vlan_map.get('use_dhcp') == 'True':
            if vlan_map.get('ve_ip'):
                raise exceptions.VirtEthCollisionConfigError(interface_num, vlan_id)
            else:
                interface_dm.ve_ips.append('dhcp')
        if not vlan_map.get('use_dhcp') or vlan_map.get('use_dhcp') == 'False':
            if not vlan_map.get('ve_ip'):
                raise exceptions.VirtEthMissingConfigError(interface_num, vlan_id)
            interface_dm.ve_ips.append(validate_partial_ipv4(vlan_map.get('ve_ip')))
        interface_dm.tags.append(vlan_id)
    interface_dm.interface_num = interface_num

    return interface_dm


def validate_interface_vlan_map(hardware_device):
    device_network_map = []
    for device_id, device_obj in hardware_device.get('interface_vlan_map').items():
        device_map = data_models.DeviceNetworkMap(device_obj.get('vcs_device_id'))
        if device_obj.get('ethernet_interfaces'):
            for eth in device_obj.get('ethernet_interfaces'):
                device_map.ethernet_interfaces.append(convert_interface_to_data_model(eth))
        if device_obj.get('trunk_interfaces'):
            for trunk in device_obj.get('trunk_interfaces'):
                device_map.trunk_interfaces.append(convert_interface_to_data_model(trunk))
        if device_obj.get('mgmt_ip_address'):
            device_map.mgmt_ip_address = device_obj.get('mgmt_ip_address')
        device_network_map.append(device_map)
    validate_vcs_device_info(device_network_map)
    return device_network_map


def get_natpool_addr_list(nat_flavor):
    addr_list = []
    if type(ip_address(nat_flavor['start_address'])) is IPv6Address:
        sh, sl = struct.unpack(">QQ",
                               socket.inet_pton(socket.AF_INET6, nat_flavor['start_address']))
        eh, el = struct.unpack(">QQ",
                               socket.inet_pton(socket.AF_INET6, nat_flavor['end_address']))
        while sl <= el:
            addr_list.append(socket.inet_ntop(socket.AF_INET6, struct.pack(">QQ", sh, sl)))
            sl += 1
    else:
        start = (struct.unpack(">L", socket.inet_aton(nat_flavor['start_address'])))[0]
        end = (struct.unpack(">L", socket.inet_aton(nat_flavor['end_address'])))[0]
        while start <= end:
            addr_list.append(socket.inet_ntoa(struct.pack(">L", start)))
            start += 1
    return addr_list


def get_loadbalancer_flavor(loadbalancer):
    flavor_repo = repo.FlavorRepository()
    flavor_profile_repo = repo.FlavorProfileRepository()
    flavor = {}
    flavor_id = loadbalancer.flavor_id
    if flavor_id:
        flavor = flavor_repo.get(db_apis.get_session(), id=flavor_id)
        if flavor and flavor.flavor_profile_id:
            flavor_profile = flavor_profile_repo.get(db_apis.get_session(),
                                                     id=flavor.flavor_profile_id)
            flavor_data = json.loads(flavor_profile.flavor_data)
            return flavor_data


def network_with_ipv6_subnet(network):
    network_driver = get_network_driver()
    for subnet_id in network.subnets:
        subnet = network_driver.get_subnet(subnet_id)
        if subnet.ip_version == 6:
            return True
    return False


def get_ipv6_address_from_conf(address_list, subnet_id, amp_network=False):
    final_address_list = []
    for address in address_list:
        address = json.loads(address)
        for key in address:
            if key == subnet_id:
                final_address_list.append({'ipv6-addr': address[key]})
    if amp_network:
        return final_address_list

    return final_address_list


def get_network_ipv6_address_from_conf(address_list, network):
    addr_list = []
    network_driver = get_network_driver()

    addr_list = get_ipv6_address_from_conf(address_list, network.id, True)
    if addr_list:
        return addr_list

    for subnet_id in network.subnets:
        subnet = network_driver.get_subnet(subnet_id)
        if subnet.ip_version == 6:
            new_addr_list = get_ipv6_address_from_conf(address_list, subnet_id, True)
            addr_list.extend(new_addr_list)
    return addr_list


def is_ipv6_subnet_in_conf(address_list, subnet_id):
    for address in address_list:
        address = json.loads(address)
        if address.get(subnet_id):
            return True
    return False


def get_ipv6_address(ifnum_oper, nics, address_list, loadbalancers_list):
    final_address_list = []
    has_v6_addr = False
    has_v4_addr = False
    opt_addr_only = CONF.a10_global.use_subnet_ipv6_addresses_only
    network_driver = get_network_driver()
    acos_mac_address = ifnum_oper['ethernet']['oper']['mac'].replace(".", "")
    for nic in nics:
        network = network_driver.get_network(nic.network_id)
        if network_with_ipv6_subnet(network) is False:
            continue

        neutron_port = network_driver.get_port(nic.port_id)
        port_mac_address = neutron_port.mac_address.replace(":", "")
        LOG.debug("[IPv6 Addr] get_ipv6_address get addr for mac:%s", port_mac_address)
        if port_mac_address == acos_mac_address:
            final_address_list = get_network_ipv6_address_from_conf(address_list, network)
            if final_address_list:
                has_v6_addr = True
            for fixed_ip in neutron_port.fixed_ips:
                if type(ip_address(fixed_ip.ip_address)) is IPv6Address:
                    if opt_addr_only and is_ipv6_subnet_in_conf(address_list, fixed_ip.subnet_id):
                        continue
                    addr_subnet = network_driver.get_subnet(fixed_ip.subnet_id)
                    subnet_prefix, subnet_mask = addr_subnet.cidr.split('/')
                    final_addr = fixed_ip.ip_address + '/' + subnet_mask
                    final_address_list.append({'ipv6-addr': final_addr})
                    LOG.debug("[IPv6 Addr] got: %s for mac: %s", final_addr, port_mac_address)
                    has_v6_addr = True
                if not has_v4_addr and type(ip_address(fixed_ip.ip_address)) is IPv4Address:
                    has_v4_addr = True
            break
    return final_address_list, (has_v6_addr and has_v4_addr)


def set_dual_stack(ifnum_address, ifnum):
    ifnum_address[(ifnum + a10constants.DUAL_STACK_MASK)] = True


def is_dual_stack(ifnum_address, ifnum):
    if (ifnum + a10constants.DUAL_STACK_MASK) in ifnum_address:
        return True
    return False


def get_acos_parameter_for_vrid(ip_version, partition):
    if ip_version == 6 and partition == "shared":
        IP_address = a10constants.IPV6_ADDRESS
        IP_address_cfg = a10constants.IPV6_ADDRESS_CFG
    elif ip_version == 6 and partition != "shared":
        IP_address = a10constants.IPV6_ADDRESS_PARTITION
        IP_address_cfg = a10constants.IPV6_ADDRESS_PARTITION_CFG
    elif ip_version == 4 and partition == "shared":
        IP_address = a10constants.IP_ADDRESS
        IP_address_cfg = a10constants.IP_ADDRESS_CFG
    else:
        IP_address = a10constants.IP_ADDRESS_PARTITION
        IP_address_cfg = a10constants.IP_ADDRESS_PARTITION_CFG
    return IP_address, IP_address_cfg
